#!/usr/bin/python
"""
Groups AMI datasets by pointing direction,
then dumps them in JSON format.
"""
from __future__ import print_function
import json
import argparse
import sys
import driveami

import logging
logging.basicConfig(level=logging.DEBUG)


def handle_args():
    """Outputs a file in JSON format listing AMI files, grouped by pointing."""

    parser = argparse.ArgumentParser(description="Filter listings for raw AMI data")

    parser.add_argument('listings',
                       help="Path to full-list (all datasets) input file")

    parser.add_argument('match',
                        help="String to match in observation groups.")

    parser.add_argument('-o', '--outfile', default=None,
                       help="Specify path to matching-datasets-list output file.")

    args = parser.parse_args()
    return args

def main():
    options = handle_args()
    with open(options.listings) as f:
        all_datasets, _ = driveami.load_listing(f,
                                 expected_datatype=driveami.Datatype.ami_la_raw)

    matching_datasets={}
    for grp_name, grp_info in all_datasets.iteritems():
        for fname in grp_info['files']:
            if str.upper(options.match) in str.upper(str(fname)):
                matching_datasets[grp_name]=grp_info
                break

    if len(matching_datasets)==0:
        print("No matches found")
        return 1

    if options.outfile is None:
        options.outfile = options.match+"_rawfiles.json"
    print('Datasets matching "{}" written to'.format(options.match),
          options.outfile)
    with open(options.outfile, 'w') as f:
        driveami.save_rawfile_listing(matching_datasets, f)


    return 0

if __name__ == "__main__":
    sys.exit(main())



"""
Utility routines for loading and saving lists of datafiles (raw or calibrated).
"""
from __future__ import absolute_import
import json
import driveami.keys as keys

class Datatype:
    magic_key = '#DATATYPE'
    ami_la_raw='AMILA_RAWFILES'
    ami_la_calibrated='AMILA_CALIBRATED_UVFITS'


datetime_format = '%Y-%m-%d %H:%M:%S'
def make_serializable(file_info_dict):
    """Returns a JSON serializable version of a file info dictionary.

    E.g. the dict returned by the `process_rawfile` routine.
    """
    d = file_info_dict.copy()
    # UTC datetime
    d[keys.time_ut] = [t.strftime(datetime_format) for t in d[keys.time_ut]]
    return d

def save_rawfile_listing(raw_obs_groups_dict, filepointer):
    savedict = raw_obs_groups_dict.copy()
    savedict[Datatype.magic_key] = Datatype.ami_la_raw
    json.dump(savedict, filepointer,
                  sort_keys=True, indent=4)

def save_calfile_listing(list_of_calobs, filepointer,
                         keep_rawtext=False):
    savedict = list_of_calobs.copy()
    if not keep_rawtext:
        for obs_id, obs_dict in list_of_calobs.iteritems():
            obs_dict.pop(keys.raw_obs_text, None)
    savedict[Datatype.magic_key] = Datatype.ami_la_calibrated
    json.dump(savedict, filepointer,
              sort_keys=True, indent=4)


def load_listing(filepointer, expected_datatype=None):
    """
    Load a json listing tagged with a driveami Datatype key.

    Args:
        filepointer: Filestream for reading.
        expected_datatype: If defined, this will raise a ValueError if
            a non-matching datatype key-entry is found.

    Returns:
        Tuple (listing_dict, found_datatype): 2-Tuple of results and datatype.


    """
    listing = json.load(filepointer)
    if (Datatype.magic_key not in listing):
        raise ValueError(
            "{} does not appear to be an AMI listing".format(filepointer)
        )
    found_datatype = listing[Datatype.magic_key]
    if expected_datatype is not None:
        if found_datatype != expected_datatype:
            raise ValueError(
            "{} does not appear to be an AMI listing of type {}".format(
                filepointer, expected_datatype)
            )
    listing.pop(Datatype.magic_key)
    return listing, found_datatype



"""
Provides a pythonic wrapper about the AMI reduce pipeline.

(Uses an instance of the interactive `reduce` program driven via pexpect.)

NB It may be possible to do this more directly by writing python wrappers about
the underlying fortran code, but this is a reasonably good quick solution.
"""

# NB I have adopted the convention that each function leaves the
# spawned instance in a 'ready to receive' state.
# So the last call should usually be to 'expect(blah)'

# Calls to ami often finish with a '\' (backslash),
# which means 'use defaults'.
# Unfortunately, this is also the python string escape character.
# So I often use raw strings to make it clear what is being sent.
from __future__ import absolute_import, print_function
import os
import shutil
import pexpect
from collections import defaultdict, namedtuple
import logging
from astropy.coordinates import SkyCoord, Longitude, Latitude
import astropy.units
import warnings
import datetime

from numpy import median

from driveami.environments import init_ami_env
import driveami.scripts as scripts

RaDecPair = namedtuple('RaDecPair', 'ra dec')

import driveami.keys as keys

logger = logging.getLogger(__name__)

def ensure_dir(dirname):
    if not os.path.isdir(dirname):
        os.makedirs(dirname)


class Reduce(object):
    """Class to provide an interface to AMI-reduce package"""
    prompt = 'AMI-reduce>'

    def __init__(self,
                 ami_rootdir,
                 array='LA',
                 working_dir='/tmp',
                 additional_env_variables=None,
                 timeout=120,
                 ):
        """
        Spawn an AMI-REDUCE instance.

        This will automatically update the list of available files,
        but will not load the full info for each file, as this is time consuming.
        (See :py:func:`load_obs_info`.)

        """
        if len(ami_rootdir) > 16:
            warnings.warn("Long AMI root path detected - this may cause bugs!\n"
                          "It is recommended to use a short symlink instead.\n")
        if working_dir is None:
            working_dir = ami_rootdir
        if not os.access(os.path.join(ami_rootdir, 'bin', 'reduce'), os.R_OK):
            raise IOError("Cannot access ami-reduce binary at: " +
                               os.path.join(ami_rootdir, 'bin', 'reduce'))
        self.working_dir = working_dir
        ami_env = init_ami_env(ami_rootdir)
        if additional_env_variables is not None:
            ami_env.update(additional_env_variables)
        self.child = pexpect.spawn('tcsh -c reduce',
                          cwd=self.working_dir,
                          env=ami_env,
                          timeout=timeout)
        self.child.expect(self.prompt)
        # Records all known information about the fileset.
        # Each file entry is initialised to a ``defaultdict(lambda : None)``
        # So if we attempt to access an unknown file attribute we get a sensible
        # answer rather than an exception.
        self.files = dict()
        # Used for updating the relevant record in self.files, also logging:
        self.active_file = None

        if array == 'LA':
            self.switch_to_large_array()
        elif array != 'SA':
            raise ValueError("Initialisation error: Array must be 'LA' or 'SA'.")
        self.array = array
        self.update_files()

        self.file_log = None
        self.file_cmd_log = None

    def switch_to_large_array(self):
        """NB resets file list"""
        p = self.child
        p.sendline('set def la')
        self.files = dict()
        p.expect(self.prompt)

    def update_files(self):
        p = self.child
        p.sendline(r'list files \ ')
#        p.sendline(r'list comment \ ')
        p.expect(self.prompt)
        # First line in 'before' is command.
        # second line is blank
        # last 4 lines are blanks and 'total obs time'
        file_lines = p.before.split('\n')[2:-4]
        for l in file_lines:
            l = l.strip('\r').strip(' ')
            cols = l.split(' ', 1)
            fname = cols[0]
            if fname not in self.files:
                self.files[fname] = {}

        p.sendline(r'list comment \ ')
#        p.sendline(r'list comment \ ')
        p.expect(self.prompt)
        file_lines = p.before.split('\n')[2:]
        for l in file_lines:
            l = l.strip('\r').strip(' ')
            cols = l.split(' ', 1)
            fname = cols[0]
            if fname in self.files:
                if len(cols) > 1:
                    self.files[fname][keys.comment] = cols[1]

    def get_obs_details(self, filename, incomplete=False):
        p = self.child
        if not incomplete:
            p.sendline(r'list observation {0} \ '.format(filename))
            p.expect(self.prompt)
        else: #incomplete observation, load fully to check end-timetamp:
            if not self.active_file==filename:
                self.set_active_file(filename)
            p.sendline(r'show observation \ ')
            p.expect(self.prompt)
        obs_lines = p.before.decode('ascii').split('\n')[2:]
        info = self.files[filename]
        if incomplete:
            warnings_dict = info.setdefault(keys.warnings, {})
            warnings_dict[keys.warning_incomplete]=True

        info[keys.raw_obs_text] = p.before
        hms_dms = Reduce._parse_coords(filename, obs_lines)
        info[keys.pointing_hms_dms] = hms_dms
        info[keys.pointing_degrees] = Reduce._convert_to_decimal_degrees(hms_dms)
        info[keys.calibrator] = Reduce._parse_calibrator(obs_lines)
        info[keys.field] = Reduce._parse_field(obs_lines)
        info.update(Reduce._parse_obs_datetime(obs_lines))
        if (info[keys.duration]==0.0 and not incomplete):
            logger.warn(
                "Incomplete obs ({}), pulling timestamps from filedata".format(
                    filename))
            self.get_obs_details(filename,incomplete=True)

        return info

    @staticmethod
    def _parse_calibrator(obs_listing):
        for line in obs_listing:
            if 'with calibrator' in line:
                tokens = line.split()
                return tokens[-1]


    @staticmethod
    def _parse_field(obs_listing):
        for line in obs_listing:
            if 'field observation' in line:
                tokens = line.split()
                return tokens[0]

    @staticmethod
    def _parse_times(line):
        """Returns (UTC time, sidereal time)"""
        tokens = line.split()
        format = '%H.%M.%S'
        ut = datetime.datetime.strptime(tokens[3], format).time()
        st = tokens[5].replace('.', ':')
        mjd = float(tokens[-2])
        return ut, st, mjd

    @staticmethod
    def _parse_obs_datetime(obs_listing):
        timeinfo = {}
        for idx, line in enumerate(obs_listing):
            if 'Tracking ' in line:
                date_str = obs_listing[idx + 1].split()[-1]
                d0 = datetime.datetime.strptime(date_str, '%d/%m/%Y')
            if 'Start time' in line:
                ut0, st0, mjd0 = Reduce._parse_times(line)
            if 'Stop time' in line:
                ut1, st1, mjd1 = Reduce._parse_times(line)

        timeinfo[keys.time_st] = (st0, st1)
        timeinfo[keys.time_mjd] = (mjd0, mjd1)
        d0 = datetime.datetime.combine(d0, ut0)
        d1 = datetime.datetime.combine(d0, ut1)
        if d1 < d0:  # Crossed midnight
            d1 = d1 + datetime.timedelta(days=1)
        timeinfo[keys.time_ut] = (d0, d1)
        duration = d1 - d0
        timeinfo[keys.duration] = duration.total_seconds() / 3600.
        return timeinfo


    @staticmethod
    def _parse_coords(filename, obs_listing):
        for line in obs_listing:
            if 'Tracking' in line:
                if not 'J2000' in line:
                    logging.warn("Obs pointing may not be in J2000 format:"
                         + filename + ", co-ord conversion may be incorrect.")

                coords_str = line[len('Tracking    : '):]
                coords_str = coords_str.strip()
                coords_str = coords_str[:-len('J2000')].strip()
                # Two cases depending whether declination is +ve or -ve:
                if '-' in coords_str:
                    ra_dec = coords_str.split('  ')
                else:
                    ra_dec = coords_str.split('   ')
                pointing = RaDecPair(ra_dec[0].replace(' ', ':'),
                                     ra_dec[1].replace(' ', ':'))
                return pointing
        raise ValueError("Parsing error for file: %s, coords not found"
                            % filename)

    @staticmethod
    def _convert_to_decimal_degrees(hms_dms_pair):
        """
        Args:

          - a tuple-pair of ('h:m:s','d:m:s') strings representing ra/dec
        """
        ra = Longitude(hms_dms_pair.ra, unit = astropy.units.hourangle)
        dec = Latitude(hms_dms_pair.dec, unit = astropy.units.deg)
        return RaDecPair(ra.degree, dec.degree)


    def load_obs_info(self):
        logger.info("Loading observation information, patience...")
        self.update_files()
        for filename, info in sorted(self.files.items()):
            if info.get(keys.pointing_degrees,None) is None:
                logger.debug("Getting obs info for %s", filename)
                try:
                    self.get_obs_details(filename)
                except Exception as error:
                    logger.exception("\n"
                                 "**********************\n"
                                 "Warning! Threw an exception trying to parse "
                                 "details for %s"
                                 "***********************\n"
                                 "\n",  filename)


    def group_obs_by_target_id(self):
        """
        Group together datasets by target id.

        Where 'target id' is everything in the filename, up to the last '-'

        Returns:
        Nested dict with structure:
        { Field name:
            {
            files: [ <list of files>],
            median_pointing: <string representation of group pointing>
            },
            ...
        }
        """

        target_groups = {}

        #Set up empty lists first:
        for filename, info in self.files.iteritems():
            target_id = filename.rsplit('-',1)[0]
            target_groups[target_id]={}
            target_groups[target_id][keys.files]=[]

        #Now put the files in
        for filename, info in self.files.iteritems():
            target_id = filename.rsplit('-',1)[0]
            target_groups[target_id][keys.files].append(filename)

        for target_id in target_groups:
            ra_list, dec_list = [],[]
            for filename in target_groups[target_id][keys.files]:
                info = self.files[filename]
                obs_pointing = info.get(keys.pointing_degrees, None)
                if obs_pointing is not None:
                    ra, dec = obs_pointing.ra, obs_pointing.dec
                    ra_list.append(ra)
                    dec_list.append(dec)
            if ra_list:
                median_ra, median_dec = median(ra_list), median(dec_list)
                target_groups[target_id][keys.target_pointing_deg] = (median_ra,median_dec)

        return target_groups


    def group_target_ids_by_pointing(self,
                                 target_id_groups,
                                 pointing_tolerance_in_degrees=0.5):
        """
        Attempt to group together datasets by inspecting pointing target.

        Returns:
        Nested dict with structure:
        { TARGET_ID:
            {
            files: [ <list of files>],
            pointing: <string representation of group pointing>
            },
            ...
        }
        """

        def _find_close_targets(first_target_id, ungrouped, skycoords_cache,
                               tolerance_deg):
            """
            Grow a network graph to include all points separated by less than
            'tolerance_deg'.

            Given a starting point ('first_target_id', and a bunch of
            un-matched positions ('ungrouped'), work through checking for
            close positions. Once a new network node has been added,
            re-run to see if this puts us close enough to another point.
            """
            logger.debug("Finding observations near "+first_target_id+" ... ")
            cluster_ids = [first_target_id]

            # Take the list of positions belonging to the current group
            # Split it into three lists, to avoid modifying the list as we
            # iterate over it:
            # Cluster positions we've already scanned for matches previously
            processed = []
            # Positions we're about to scan for matches:
            process_next = [skycoords_cache[first_target_id]]
            # Newly added positions that could extend the cluster in the next
            # iteration:
            newly_included = []

            while process_next:
                for pointing1 in process_next:
                    for id2 in list(ungrouped):
                        pointing2 = skycoords_cache[id2]
                        if pointing1.separation(pointing2).degree < tolerance_deg:
                            cluster_ids.append(id2)
                            newly_included.append(pointing2)
                            ungrouped.remove(id2)
                            logger.debug("... " + id2 + " added to group.")
                # Cycle the lists before next iteration
                processed.extend(process_next)
                process_next = newly_included
                newly_included = []
            logger.debug("... {} rawfiles in group.".format(len(cluster_ids)))
            return cluster_ids


        ungrouped = set(target_id_groups.keys())
        # Cache a pre-built set of Astropy SkyCoord objects, one for each id
        # This avoids constantly rebuilding as we iterate through.
        temp_skycoords = {}
        for id in ungrouped:
            ra_dec = target_id_groups[id][keys.target_pointing_deg]
            temp_skycoords[id] =SkyCoord(ra_dec[0], ra_dec[1], unit='deg')

        pointing_groups_dict = {}

        #Now, pop off un-searched IDs and match them up, one-by-one until done
        while ungrouped:
            cluster_ids = _find_close_targets(ungrouped.pop(), ungrouped,
                                              temp_skycoords,
                                              pointing_tolerance_in_degrees)
            first_id = sorted(cluster_ids)[0]
            pointing_groups_dict[first_id] = {}
            pointing_groups_dict[first_id][keys.target_pointing_deg]=(
                target_id_groups[first_id][keys.target_pointing_deg])
            pointing_groups_dict[first_id][keys.files]=[]
            for target_id in cluster_ids:
                pointing_groups_dict[first_id][keys.files].extend(
                    target_id_groups[target_id][keys.files])

        return pointing_groups_dict


    def close_per_file_logs(self):
        """Close any logging file handlers from the last file"""
        if self.file_log is not None:
            for hdlr in self.file_log.handlers:
                if hasattr(hdlr, 'close'):
                    hdlr.close()
        if self.file_cmd_log is not None:
            for hdlr in self.file_cmd_log.handlers:
                if hasattr(hdlr, 'close'):
                    hdlr.close()

    def _setup_file_loggers(self, filename, file_logdir):
        target = os.path.splitext(filename)[0]

        self.close_per_file_logs()

        # It does no harm to set up the loggers,
        # irrespective of whether we write them to file-
        # The calling code could potentially grab them for other uses.

        self.file_log = logging.getLogger('.'.join((logger.name, target)))
        self.file_log.propagate = False
        self.file_log.setLevel(logging.DEBUG)
        self.file_cmd_log = logging.getLogger(
                              '.'.join((logger.name, 'commands', target)))
        self.file_cmd_log.propagate = False
        self.file_cmd_log.setLevel(logging.DEBUG)

        if file_logdir is not None:
            ensure_dir(file_logdir)
            fh = logging.FileHandler(
                         os.path.join(file_logdir, target + '.ami.log'),
                         mode='w')
            self.file_log.addHandler(fh)

            fh = logging.FileHandler(
                         os.path.join(file_logdir, target + '.ami.commands'),
                         mode='w')
            self.file_cmd_log.addHandler(fh)
        else:
            self.file_log.addHandler(logging.NullHandler())
            self.file_cmd_log.addHandler(logging.NullHandler())

    def run_command(self, command):
        self.file_cmd_log.debug(command)
        self.child.sendline(command)
        self.child.expect(self.prompt)
        self.file_log.debug('%s%s', self.prompt, self.child.before)
        self._parse_command_output(command, self.child.before.split('\n'))
        return self.child.before.split('\n')

    def _parse_command_output(self, command, output_lines):
#        try:
        file_info = self.files[self.active_file]
        if 'apply rain' in command:
            rain_amp_corr = self._parse_rain_results(output_lines)
            file_info[keys.rain] = rain_amp_corr
#            logger.info("Rain mean amplitude correction factor: %s",
#                             rain_amp_corr)
        if 'flag' in command:
            flagging = self._parse_flagging_results(output_lines)
            file_info[keys.flagged_max] = max(flagging,
                                          file_info.get(keys.flagged_max,None))

        if 'reweight' in command:
            est_noise = self._parse_reweight_results(output_lines)
            file_info[keys.est_noise_jy] = est_noise
#            logger.info("Estimated noise: %s mJy", est_noise * 1000.0)
                # self.files[self.active_file][keys.flagging_max]

        if 'cal inter' in command:
            avail, daysep = self._parse_cal_inter_results(output_lines)
            file_info[keys.archive_cal_available]=avail
            file_info[keys.archive_cal_days_apart]=daysep

#        except Exception as e:
#            raise ValueError("Problem parsing command output for file: %s,",
#                             "command: %s, error message:\n%s"
#                             ,self.active_file, command, e.msg)


    def _parse_rain_results(self, output_lines):
        for line in output_lines:
            if "Mean amplitude correction factor" in line:
                return float(line.strip().split()[-1])
        raise ValueError("Parsing error, could not find rain modulation.")

    def _parse_flagging_results(self, output_lines):
        for line in output_lines:
            if "samples flagged" in line:
                if "Total of" in line:
                    tokens = line.strip().split()
                    for t in tokens:
                        if '%' in t:
                            return float(t.strip('%'))

    def _parse_reweight_results(self, output_lines):
        for line in output_lines:
            if "estimated noise" in line:
                tokens = line.strip().split()
                return float(tokens[-2])
        raise ValueError("Parsing error, could not find noise estimate.")

    def _parse_cal_inter_results(self, output_lines):
        days_apart=None
        archive_cal_available=True
        for line in output_lines:
            if 'fluxes for this source not available' in line:
                archive_cal_available=False
            if 'days apart' in line:
                tokens = line.strip().split()
                days_apart=float(tokens[6])
        return archive_cal_available, days_apart

    def run_script(self, script_string):
        """Takes a script of commands, one command per line"""
        command_list = script_string.split('\n')
        for command in command_list:
            self.run_command(command)


    def set_active_file(self, filename, file_logdir=None):
        filename = filename.strip()  # Ensure no stray whitespace
        self.active_file = filename
        self._setup_file_loggers(filename, file_logdir)
        file_listing = self.run_command(r'file %s \ ' % filename)
        incomplete_obs = False
        for line in file_listing:
            if 'incomplete observation' in line:
                incomplete_obs = True

        if incomplete_obs:
            self.get_obs_details(filename, incomplete=True)
        else:
            self.get_obs_details(filename, incomplete=False)
        logger.debug('Active file: %s', filename)


    def write_files(self, rawfile, output_dir,
                    write_command_template = scripts.write_command,
                    write_command_overrides=None):
        """Writes out UVFITs files.

        NB: You should use this rather than performing writes manually:
        ``reduce`` cannot handle long file paths,
        so rather than cripple the scripting functionality,
        this function hacks around the limitations.
        Kludgey but effective.
        """
        ensure_dir(output_dir)
        tgt_name = os.path.splitext(rawfile)[0]
        tgt_path = os.path.join(output_dir, tgt_name + '.fits')
        if self.files[rawfile][keys.calibrator] is not None:
            cal_basename = (tgt_name+'_cal_'+
                            self.files[rawfile][keys.calibrator]+'.fits')
            cal_path = os.path.join(output_dir, cal_basename)
        else:
            cal_path=None
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            tgt_temp = os.tempnam(self.working_dir, 'ami_') + '.fits'
            cal_temp = os.tempnam(self.working_dir, 'ami_') + '.fits'

        if cal_path is None:
            output_paths_string = os.path.basename(tgt_temp)
        else:
            output_paths_string = " ".join((os.path.basename(tgt_temp),
                                            os.path.basename(cal_temp)))
        logger.debug("Writing to temp files %s" % output_paths_string)

        write_command_args = scripts.write_command_defaults.copy()
        if write_command_overrides is not None:
            write_command_args.update(write_command_overrides)

        write_command_args['output_paths']=output_paths_string
        write_command = write_command_template.format(**write_command_args)
        self.run_command(write_command)

        logger.debug("Renaming tempfile %s -> %s", tgt_temp, tgt_path)
        shutil.move(tgt_temp, tgt_path)
        info = self.files[self.active_file]
        info[keys.target_uvfits] = os.path.abspath(tgt_path)
        if cal_path is not None:
            logger.debug("Renaming tempfile %s -> %s", cal_temp, cal_path)
            shutil.move(cal_temp, cal_path)
            info[keys.cal_uvfits] = os.path.abspath(cal_path)
        logger.debug("Wrote target, calib. UVFITs to:\n\t%s\n\t%s",
                         tgt_path, cal_path)

    def update_flagging_info(self):
        lines = self.run_command(r'show flagging no yes \ ')
        final_flagging = self._parse_flagging_results(lines)
        self.files[self.active_file][keys.flagged_final] = final_flagging
#        logger.info("Final flagging estimate: %s%%", final_flagging)



from __future__ import absolute_import
import logging
import os
import json
from colorlog import ColoredFormatter
import driveami.keys as keys
import driveami.scripts as scripts
from driveami.reduce import Reduce

from driveami.serialization import (Datatype, make_serializable,
                                    save_calfile_listing,save_rawfile_listing,
                                    load_listing)

logger = logging.getLogger('ami')

def ensure_dir(dirname):
    if not os.path.isdir(dirname):
        os.makedirs(dirname)


def process_rawfile(rawfile, output_dir,
                    reduce,
                    file_logging=True,
                    script=scripts.standard_reduction):
    """
    A convenience function applying sensible defaults to reduce a rawfile.

    Args:
    rawfile: Name of a file in the ami data dir, e.g. "SWIFT121101-121101.raw"
    output_dir: Folder where UVFITS for the target and calibrator will be output.
    reduce: instance of ami.Reduce
    array: 'LA' or 'SA' (Default: LA)
    script: Reduction commands.

    Returns:
        - A dictionary containing information about the rawfile,
          e.g. pointing, calibrator name, rain modulation.
          See also: ``ami.keys``
    """
    r = reduce
    if file_logging:
        file_logdir = output_dir
    else:
        file_logdir = None
    r.set_active_file(rawfile, file_logdir)
    r.run_script(script)
    r.update_flagging_info()
    r.write_files(rawfile, output_dir,
                  write_command_overrides={'channels':'3-8'})
    r.files[rawfile][keys.obs_name] = os.path.splitext(rawfile)[0]
    info_filename = os.path.splitext(rawfile)[0] + '.json'
    with open(os.path.join(output_dir, info_filename), 'w') as f:
        json.dump(make_serializable(r.files[rawfile]), f,
                  sort_keys=True, indent=4)
    return r.files[rawfile]


def get_color_log_formatter():
    date_fmt = "%y-%m-%d (%a) %H:%M:%S"
    color_formatter = ColoredFormatter(
            "%(log_color)s%(asctime)s:%(levelname)-8s%(reset)s %(blue)s%(message)s",
            datefmt=date_fmt,
            reset=True,
            log_colors={
                    'DEBUG':    'cyan',
                    'INFO':     'green',
                    'WARNING':  'yellow',
                    'ERROR':    'red',
                    'CRITICAL': 'red',
            }
    )
    return color_formatter

def get_color_stdout_loghandler(level):
    stdout_loghandler = logging.StreamHandler()
    stdout_loghandler.setFormatter(get_color_log_formatter())
    stdout_loghandler.setLevel(level)
    return stdout_loghandler


#!/usr/bin/python

"""
Groups AMI datasets by pointing direction,
then dumps them in JSON format.
"""
import argparse
import sys
import logging

import driveami
from driveami.environments import default_ami_dir

logger=logging.getLogger()
logger.setLevel(logging.DEBUG)
logger.addHandler(driveami.get_color_stdout_loghandler(logging.DEBUG))

def handle_args():
    default_array = 'LA'
    default_full_listings_filename = 'all_ami_rawfiles'

    parser = argparse.ArgumentParser(
        description="Generate listings for raw AMI data")

    parser.add_argument("--ami", default=default_ami_dir,
                        help="Path to AMI directory, default: " + default_ami_dir)

    parser.add_argument("--array", default=default_array,
                        help="Array data to work with (SA/LA), defaults to: "
                             + default_array)

    parser.add_argument('-o', '--outfile',
                        default=default_full_listings_filename,
                        help="Path to output file (NB, will be suffixed "
                             "to create two versions) "
                             "default: "
                             + default_full_listings_filename)

    args = parser.parse_args()
    logger.info("Will output listings of all datasets to '{}'".format(
                                                                args.outfile))
    return args


def main():
    options = handle_args()
    grouped_by_id_filename = options.outfile + '_by_id.json'
    grouped_by_pointing_filename = options.outfile + '_by_pointing.json'
    metadata_filename = options.outfile + '_metadata.json'

    r = driveami.Reduce(options.ami, options.array)
    logger.info("Loading observation metadata.")
    r.load_obs_info()
    logger.info("Grouping observations by target ID")
    id_groups = r.group_obs_by_target_id()
    with open(grouped_by_id_filename, 'w') as f:
        driveami.save_rawfile_listing(id_groups, f)
    logger.info("Grouping targets by pointing")
    pointing_groups = r.group_target_ids_by_pointing(id_groups,
                                                     pointing_tolerance_in_degrees=0.5)
    with open(grouped_by_pointing_filename, 'w') as f:
        driveami.save_rawfile_listing(pointing_groups, f)

    with open(metadata_filename, 'w') as f:
        rawfile_dict = { fname:driveami.make_serializable(info) for fname, info
                         in r.files.iteritems()}
        driveami.save_rawfile_listing(rawfile_dict, f)

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""Defines various shell environments for use with subprocess.

We generally grab the current shell env, then update paths with package specific
directories.

This is a nice way to ensure everything happens in its own sandbox,
without confusing interaction between conflicting variables.

Note the utility functions are overkill for the AMI package, but
are handy when defining multiple environments.
"""
import os

default_ami_dir = os.environ.get('AMI_DIR',os.path.expanduser("~/ami"))
default_output_dir = os.environ.get('AMI_RESULTS',
                                   os.path.expanduser("~/ami_results"))

def _append_to_path(env, path, extra_dir):
    """Deal nicely with empty path, minimise boilerplating."""
    env_dirs = env.get(path, '').split(':')
    env_dirs.append(extra_dir)
    env[path] = ':'.join(env_dirs)

def _add_bin_lib_python(env, package_group_dir, include_python=True):
    """Add the bin, lib, (and optionally python-packages) subdirs of
    `package_group_dir` to the relevant environment paths."""
    _append_to_path(env, "PATH", os.path.join(package_group_dir, 'bin'))
    _append_to_path(env, 'LD_LIBRARY_PATH',
                    os.path.join(package_group_dir, 'lib'))
    if include_python:
        _append_to_path(env, 'PYTHONPATH',
                        os.path.join(package_group_dir, 'python-packages'))

def init_ami_env(ami_topdir):
    """Returns an environment dictionary configured for AMI-reduce execution.
        
    `ami_topdir` should be the top directory of the AMI-reduce installation.
    
    """
    ami_env = {}
    ami_env["AMI_DIR"] = ami_topdir
    _add_bin_lib_python(ami_env, ami_topdir, include_python=False)
    ami_env["PGPLOT_DIR"] = os.path.join(ami_topdir , "lib/pgplot")
    ami_env["PGPLOT_FONT"] = os.path.join(ami_topdir , "lib/pgplot/grfont.dat")
    ami_env["PGPLOT_RGB"] = os.path.join(ami_topdir , "lib/pgplot/rgb.txt")
    ami_env["PGPLOT_DEV"] = "/xwin"
    return ami_env



comment = 'comment'
pointing_hms_dms = 'pointing_hms_dms'
pointing_degrees = 'pointing_degrees'
calibrator = 'calibrator'
rain = 'rain_amp_corr'
flagged_max = 'flagged_max_percent'
flagged_final = 'flagged_final_percent'
est_noise_jy = 'estimated_noise_jy'
target_uvfits = 'target_uvfits'
cal_uvfits = 'calib_uvfits'
files = 'files'
field = 'field'
target_pointing_deg = 'target_median_pointing'
group_name = 'group_name'
obs_name = 'obs_name'
time_ut = 'time_utc'
time_mjd = 'time_mjd'
time_st = 'time_sidereal'
duration = 'duration_hrs'
archive_cal_available = 'archive_cal_available'
archive_cal_days_apart = 'archive_cal_days_apart'
raw_obs_text = 'raw_obs_listing_text'
warnings = 'warning_flags'
warning_incomplete = 'incomplete_observation'



standard_reduction = \
r"""
version
flag all
flag int \
flag amp field no \
flag data \
update pcal \
subtract modmeans \
subtract zeros \
subtract means \
fft \
frotate forward y n \
apply rain \
flag amp field no 0.95 1 yes \
flag amp field no 0.15 30 yes \
flag amp field yes no 60 \
frotate forward n y \
cal inter \
reweight \
show flagging no yes \
smooth 20 \
show flagging no yes \
flag bad calibrator \
flag amp field no 0.25 1 yes \
smooth 10 \
flag amp field no 0.075 1 yes \
scan dat cal yes \
scan dat field yes \
show flagging no yes \
"""

no_calibrator_reduction = \
r"""
version
flag all
flag int \
flag amp no \
flag data \
update pcal \
subtract modmeans \
subtract zeros \
subtract means \
fft \
frotate forward y n \
apply rain \
flag amp  no 0.95 1 yes \
flag amp  no 0.15 30 yes \
flag amp  yes no 60 \
frotate forward n y \
cal inter \
reweight \
show flagging no yes \
smooth 20 \
show flagging no yes \
flag bad \
flag amp no 0.25 1 yes \
smooth 10 \
flag amp  no 0.075 1 yes \
scan dat yes \
show flagging no yes \
"""
# NB second arg refers to 'update entry in observation database'
# I currently know of no use cases for that to be 'yes'.
write_command=\
r'write fits {if_severely_flagged} no {baselines} {channels} {sample_range} {output_paths} \  '

write_command_defaults = {'if_severely_flagged':'yes',
                          'baselines':'all',
                          'channels':'3-8',
                          'sample_range':'all'}


#!/usr/bin/python
from __future__ import print_function
import argparse
import os
import sys
import logging
import datetime

import driveami
from driveami.environments import (default_ami_dir, default_output_dir)

def handle_args():
    """
    Default values are defined here.
    """
    parser = argparse.ArgumentParser(
        description="Calibrate raw AMI data and produce uvFITs")
    parser.add_argument("-t", "--topdir", default=default_output_dir,
                        help="Top level data-output directory, default is : " +
                            default_output_dir)

    default_outfile = "calibrated_files.json"

    parser.add_argument('-o', '--outfile', nargs='?', default=default_outfile,
                        help='Specify filename for output listing of calibrated '
                             'data, default: '+default_outfile)

    parser.add_argument('groups_file', metavar='groups_to_process.json', nargs='?',
                        help='Specify file listing rawfiles for processing '
                             '(overrides all other file options)')

    parser.add_argument("--ami", default=default_ami_dir,
                       help="Path to AMI directory, default: " + default_ami_dir)

    parser.add_argument('-s', '--script', help='Specify non-standard reduction script')

    parser.add_argument('-f', '--files', nargs='*',
                        help='Specify individual files for reduction')

    parser.add_argument('-g', '--group', dest='groupname', default='NOGROUP',
                        help='Specify group name for individually specified files')

    # parser.add_argument('-r', '--array', default='LA',
    #                     help='Specify array (SA/LA) for individually specified files')

    options = parser.parse_args()
    options.ami_dir = os.path.expanduser(options.ami)
    options.topdir = os.path.expanduser(options.topdir)

    if options.script:
        with open(options.script) as f:
            options.script = f.read()

    if options.groups_file:
        print("Reducing files listed in:", options.groups_file)
        with open(options.groups_file) as f:
            data_groups, _ = driveami.load_listing(f,
                               expected_datatype=driveami.Datatype.ami_la_raw)
    elif options.files:
        data_groups = {options.groupname: {'files':options.files}}
    else:
        parser.print_help()
        sys.exit()

    # print "SCRIPT:", options.script
    return options, data_groups

def output_preamble_to_log(data_groups):
    logger.info("*************************************")
    logger.info("Processing with AMI reduce:\n"
                 "--------------------------------")
    for key in sorted(data_groups.keys()):
        logger.info("%s:", key)
        for f in data_groups[key][driveami.keys.files]:
            logger.info("\t %s", f)
        logger.info("--------------------------------")
    logger.info("*************************************")

def process_data_groups(data_groups, output_dir, ami_dir,
                    array='LA',
                    script=None):
    """Args:
    data_groups: Dictionary mapping groupname -> list of raw filenames
    output_dir: Folder where dataset group subfolders will be created.
    ami_dir: Top dir of the AMI ``reduce`` installation.
    array: 'LA' or 'SA' (Default: LA)
    """
    if not script:
        script = driveami.scripts.standard_reduction

    processed_files_info = {}
    for grp_name in sorted(data_groups.keys()):

        try:
            r = driveami.Reduce(ami_dir, array=array)
            files = data_groups[grp_name][driveami.keys.files]
            grp_dir = os.path.join(output_dir, grp_name, 'ami')
            driveami.ensure_dir(grp_dir)
            logger.info('Calibrating rawfiles and writing to {}'.format(grp_dir))
            for rawfile in files:
                try:
                    logger.info("Reducing rawfile %s ...", rawfile)
                    file_info = driveami.process_rawfile(rawfile,
                                        output_dir=grp_dir,
                                        reduce=r,
                                        script=script)
                except (ValueError, IOError) as e:
                    logger.exception("Hit exception reducing file: %s\n"
                                 "Exception reads:\n%s\n",
                                 rawfile, e)
                    continue
                # Also save the group assignment in the listings:
                file_info[driveami.keys.group_name] = grp_name
                processed_files_info[rawfile] = driveami.make_serializable(file_info)
        except Exception as e:
            logger.exception(
                "Hit exception (probable timeout) reducing group: {}".format(
                    grp_name))
            continue
    return processed_files_info


def main(options, data_groups):
    output_preamble_to_log(data_groups)
    processed_files_info = process_data_groups(data_groups,
                                options.topdir,
                                options.ami_dir,
                                array='LA',
                                script=options.script)

    with open(options.outfile, 'w') as f:
        driveami.save_calfile_listing(processed_files_info, f)
    sys.exit(0)


if __name__ == "__main__":
    options, data_groups = handle_args()
    timestamp = datetime.datetime.now().strftime("%y-%m-%dT%H%M%S")
    logging.basicConfig(format='%(name)s:%(levelname)s:%(message)s',
                    filemode='w',
                    filename="driveami_calibrate_log_{}".format(timestamp),
                    level=logging.DEBUG)
    logger = logging.getLogger()
    logger.addHandler(driveami.get_color_stdout_loghandler(logging.INFO))
#    logging.basicConfig(level=logging.WARN)
    main(options, data_groups)